import pprint
import sys
import os
import logging
import argparse
import datetime
import multiprocessing
from orthanc_api_client import helpers
from .helpers.scheduler import Scheduler
from .dicom_migrator import DicomMigrator, Message
import csv

from orthanc_api_client import OrthancApiClient
logger = logging.getLogger(__name__)

class IdsMigrator(DicomMigrator):
    """
    Uses the DicomMigrator to migrate a list of studies based on a csv file containing the UIDs.
    """

    def __init__(self,
                 api_client: OrthancApiClient,
                 ids_list_file_path: str,               # Path of the file containing the ids of the studies to migrate
                 source_modality: str = None,           # Source modality as configured in Orthanc (alias)
                 max_cfind_study_count: int = None,     # Known maximum amount of studies retrievable from the source modality at once
                 destination_modality: str = None,      # Destination modality as configured in Orthanc (alias)
                 destination_aet: str = None,           # Destination AET
                 delete_from_source: bool = False,      # once the data has been migrated, delete it from source (only vali
                 scheduler: Scheduler = None,
                 worker_threads_count: int = multiprocessing.cpu_count() - 1,  # by default, use all CPUs but one for compression
                 exit_on_error: bool = False
                 ):

        super().__init__(
            api_client=api_client,
            source_modality=source_modality,
            max_cfind_study_count=max_cfind_study_count,
            destination_modality=destination_modality,
            destination_aet=destination_aet,
            delete_from_source=delete_from_source,
            scheduler=scheduler,
            worker_threads_count=worker_threads_count,
            exit_on_error=exit_on_error
        )

        self._ids_list_file_path = ids_list_file_path

    def execute(self):
        super().execute()

        logger.info("Path of the file containing the ids of the studies to migrate: " + str(self._ids_list_file_path))

        # The input could be a simple text file, each line being an id but also the output of another script with more
        # information, so we take the first item of each line
        ids_list = []
        with open(self._ids_list_file_path, 'r') as csv_file:
            reader = csv.reader(csv_file)
            for line in reader:
                ids_list.append(line[0])

        for id in ids_list:
            logger.info(f"Processing id {id}")
            self.push_message(Message(dicom_id=id))

        self.stop_threads()

        logger.info("--------------------------------------------------------------------")
        logger.info("Migration completed")


# examples:
# migrator as source: MIGRATOR --> DESTINATION:  ! use the --delete_from_source option with care !!!
# python orthanc_tools/ids_migrator.py --url=http://localhost:8042 --user=user --password=pwd --destination_modality=orthanc-debug --ids_list_file_path=/ids_list.csv --run_only_at_night_and_weekend --night_start_hour=18 --night_end_hour=6


if __name__ == '__main__':
    level = logging.INFO

    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Migrate the content of a remote modality to another modality')
    parser.add_argument('--url', type=str, default=None, help='Orthanc url (migrator)')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--destination_modality', type=str, default=None, help='Destination modality (alias)')
    parser.add_argument('--destination_aet', type=str, default=None, help='Destination AET')
    parser.add_argument('--source_modality', type=str, default=None, help='Source modality (alias)')
    parser.add_argument('--ids_list_file_path', type=str,  required='IDS_LIST_FILE_PATH' not in os.environ, help='Path of the file containing the ids of the studies to migrate')
    parser.add_argument('--delete_from_source', default=False, action='store_true', help='delete data from source (only if source is an Orthanc)')
    parser.add_argument('--worker_threads_count', type=int, default=1, help='Worker threads count')
    parser.add_argument('--exit_on_error', default=False, action='store_true', help='if True, the script will exit in case of error')

    Scheduler.add_parser_arguments(parser)

    args = parser.parse_args()

    url = os.environ.get("ORTHANC_URL", args.url)
    user = os.environ.get("ORTHANC_USER", args.user)
    pwd = os.environ.get("ORTHANC_PWD", args.password)
    destination_modality = os.environ.get("DESTINATION_MODALITY", args.destination_modality)
    destination_aet = os.environ.get("DESTINATION_AET", args.destination_aet)
    source_modality = os.environ.get("SOURCE_MODALITY", args.source_modality)
    ids_list_file_path = os.environ.get("IDS_LIST_FILE_PATH", args.ids_list_file_path)
    worker_threads_count = int(os.environ.get("WORKER_THREADS_COUNT", str(args.worker_threads_count)))

    scheduler = Scheduler.create_from_args_and_env_var(args)

    if os.environ.get("DELETE_FROM_SOURCE", None) is not None:
        delete_from_source = os.environ.get("DELETE_FROM_SOURCE") == "true"
    else:
        delete_from_source = args.delete_from_source

    if os.environ.get("EXIT_ON_ERROR", None) is not None:
        exit_on_error = os.environ.get("EXIT_ON_ERROR") == "true"
    else:
        exit_on_error = args.exit_on_error

    migrator = IdsMigrator(
        api_client=OrthancApiClient(url, user=user, pwd=pwd),
        ids_list_file_path=ids_list_file_path,
        destination_modality=destination_modality,
        destination_aet=destination_aet,
        source_modality=source_modality,
        delete_from_source=delete_from_source,
        scheduler=scheduler,
        worker_threads_count=worker_threads_count,
        exit_on_error=exit_on_error
    )

    migrator.execute()



import inquirer
from pathlib import Path
import logging
import argparse
from orthanc_api_client import OrthancApiClient
import os, time, sys
import zipfile
import tempfile

logger = logging.getLogger(__name__)

'''
This script is interactive!
So, run it from the console, follow instructions and enjoy...
'''

#TODO: modify to use the folder importer!

class OrthancUploader:
    def __init__(self,
                 api_client: OrthancApiClient,
                 path: str
                 ):
        self._api_client = api_client
        self._path = path

    def get_folder(self, current_path = Path('.')):
        """ Print a numbered list of the subfolders in the working directory
        (i.e. the directory the script is run from), and returns the directory
        the user chooses.
        """
        answers = {'studies_path': 'None'}

        while answers['studies_path'] != '.':
            print("Browse to the folder you want to upload to Orthanc, then select the . to validate.")
            questions = [
                inquirer.List(
                    "studies_path",
                    message="Curent folder: " + str(current_path.absolute()),
                    choices=['.'] + ['..'] + [str(d) for d in current_path.iterdir() if d.is_dir()]
                )
            ]
            answers = inquirer.prompt(questions)
            if answers['studies_path'] == '..':
                current_path = current_path.absolute().parent
            elif answers['studies_path'] != '.':
                current_path = Path(answers['studies_path'])
        return str(current_path.absolute())

    def upload_folder_and_label(self, folder_path, labels_list):
        """
        Recursively get the files of the folder to
        - upload each file
        - apply the labels on the study
        """

        for path in os.listdir(folder_path):
            full_path = os.path.join(folder_path, path)
            if os.path.isfile(full_path):

                if zipfile.is_zipfile(full_path):
                    with tempfile.TemporaryDirectory() as tempDir:
                        with zipfile.ZipFile(full_path, 'r') as z:
                            z.extractall(tempDir)
                        self.upload_folder_and_label(folder_path=tempDir, labels_list=labels_list)
                else:
                    retry_count = 0
                    retry_delays = [5, 20, 60, 300, 900, 1800, 3600, 7200]
                    while retry_count <= 8:
                        if retry_count >= 1:
                            delay = retry_delays[retry_count - 1]
                            logger.info(f"waiting {delay} seconds before retrying the upload of {full_path}")
                            time.sleep(delay)
                        try:
                            # here, we should have only files (and no zip file)
                            instance_orthanc_ids = orthanc_client.upload_file(full_path, ignore_errors=True)
                            if len(instance_orthanc_ids) == 0:
                                logger.error(f"File not uploaded: {full_path}.")
                                break
                            study_orthanc_id = orthanc_client.instances.get_parent_study_id(instance_orthanc_ids[0])
                            orthanc_client.studies.add_labels(orthanc_id=study_orthanc_id, labels=labels_list)
                            break
                        except Exception as e:
                            retry_count += 1
                            if retry_count == 8:
                                logger.error(f"Error while uploading this file: {full_path}. Exception: {str(e)}")
                                logger.error(f"too many attempts, exiting...")
                                sys.exit(1)
                            else:
                                logger.warning(f"Error while uploading this file, retrying...: {full_path}. Exception: {str(e)}")
            elif os.path.isdir(full_path):
                self.upload_folder_and_label(folder_path=full_path, labels_list=labels_list)

    def execute(self):

        # let the user choose the folder to upload
        folder_path = self.get_folder(Path(self._path))

        # let's get the existing labels
        labels_list = orthanc_client.get_all_labels()
        orthanc_labels_chosen_list = []

        # let the user choose the labels to apply among the existing ones
        if len(labels_list)>0:
            questions_labels = [
                inquirer.Checkbox(
                    "orthanc_labels_chosen_list",
                    message="Choose the existing labels to apply",
                    choices=labels_list,
                )
            ]

            answers = inquirer.prompt(questions_labels)
            orthanc_labels_chosen_list = answers['orthanc_labels_chosen_list']

        # let the user manually add extra labels
        answers = {'orthanc_extra_label': 'None'}
        while answers['orthanc_extra_label'] != '':
            print(f"Current labels list: {orthanc_labels_chosen_list}")
            questions_labels_manual = [
                inquirer.Text(
                    "orthanc_extra_label",
                    message="Type an extra label to apply, leave it empty and press ENTER if any"
                ),
            ]

            answers = inquirer.prompt(questions_labels_manual)
            orthanc_extra_label = answers['orthanc_extra_label']
            if orthanc_extra_label != "":
                orthanc_labels_chosen_list.append(orthanc_extra_label)

        print("-----------------------------------------------------------------------------------")
        print(f"Folder to import: {folder_path}")
        print(f"Labels to apply: {orthanc_labels_chosen_list}")
        questions_continue = [
            inquirer.Confirm("continue", message="Ready to start?", default=True)
        ]

        answers = inquirer.prompt(questions_continue)
        if answers['continue'] == False:
            print("Aborted!")
            exit(-1)

        logger.info(f"starting import of folder {folder_path} ...")

        self.upload_folder_and_label(folder_path=folder_path, labels_list=orthanc_labels_chosen_list)

        logger.info("End of upload!")


if __name__ == '__main__':
    level = logging.INFO

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description="Upload studies from a selected folder and recursively in all subfolders to an Orthanc and apply labels.")
    parser.add_argument('--url', help="The url of the Orthanc to upload studies to.", default="http://localhost:8042")
    parser.add_argument('--user', help="The user of the Orthanc to upload studies to.", default="demo")
    parser.add_argument('--password', help="The password of the Orthanc to upload studies to.", default="demo")
    parser.add_argument('--api_key', type=str, default=None, help='Orthanc api-key')
    parser.add_argument('--start_path', help="The path used as start path for the selection of the folder containing studies to upload. Default value: script execution path", default=".")

    args = parser.parse_args()

    orthanc_url = args.url
    orthanc_user = args.user
    orthanc_password = args.password
    orthanc_api_key = args.api_key
    start_path = args.start_path

    orthanc_client = None
    if orthanc_api_key is not None:
        orthanc_client=OrthancApiClient(orthanc_url, headers={"api-key":orthanc_api_key})
    else:
        orthanc_client=OrthancApiClient(orthanc_url, user=orthanc_user, pwd=orthanc_password)

    uploader = OrthancUploader(api_client=orthanc_client, path=start_path)

    uploader.execute()



import contextlib
import typing
# from hl7Lib import Hl7MessageParser
from .hl7_message_parser import Hl7MessageParser


class Hl7ReportParser(Hl7MessageParser):

    def __init__(self, specific_fields: dict = None):
        super(Hl7ReportParser, self).__init__()

        # Let's add "standards" field
        self.add_fields_definitions({
            # --- PID segment
            'PatientID': 'PID.F3.R1.C1',
            'PatientName': 'PID.F5',  # IHE recommend to use the whole Field 5 as the patient name (it's usually full of ^ that is a separator in HL7 so we must take it as is without trying to parse it !
            'PatientBirthDate': 'PID.F7',

            # --- OBR segment
            'StudyInstanceUID': 'OBR.F3.R1.C1',

            # --- OBX Segment
            'Base64Report': 'OBX.F5.R1.C5'
        })

        # Let's add specific fields (they will override the default ones if any)
        if specific_fields is not None:
            self.add_fields_definitions(specific_fields)

    def parse(self, hl7_message: str) -> typing.Dict:

        # extract field the default way
        values = super(Hl7ReportParser, self).parse(hl7_message, strict = False)

        # keep only the first 5 components of the name according to http://dicom.nema.org/dicom/2013/output/chtml/part05/sect_6.2.html (check PN VR definition)
        values['PatientName'] = '^'.join(values['PatientName'].split('^')[:5])

        # clean birthdate
        values['PatientBirthDate'] = values['PatientBirthDate'][0:8]

        return values




from optparse import OptionParser

import hl7
import typing
import os.path
import six
import socket
import sys
from .hl7_message_validator import Hl7MessageValidator


class MLLPClient:
    """
    A basic, blocking, HL7 MLLP client based upon :py:mod:`socket`.

    Can be used by the ``with`` statement to ensure :py:meth:`MLLPClient.close`
    is called::

        with MLLPClient(host, port) as client:
            client.send(msg)
    """
    def __init__(self, host: str, port: int, encoding: str = 'iso-8859-1'):
        self.encoding = encoding
        self.sb = b'\x0b' # <SB>, vertical tab
        self.eb = b'\x1c' # <EB>, file separator
        self.cr = b'\r' # <CR>, \r
        self._extractor = Hl7MessageValidator(sb = self.sb, eb = self.eb, cr = self.cr, encoding = encoding)

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((host, port))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, trackeback):
        self.close()

    def close(self):
        """Release the socket connection"""
        self.socket.close()

    def _receive(self):
        end_seq = self.eb + self.cr  # end sequence for the MLLP message
        try:
            # receive the first 3 char of the byte stream (an MLLP HL7 message must at least be 3 chars long (the sb, eb and cr)
            bline = self.socket.recv(3)
        except socket.timeout:
            self.request.close()
            return # TODO: throw exception on errors

        if len(bline) == 0: # connection has been closed
            return # TODO: throw exception on errors

        # check the first char of the MLLP message
        if bline[0] != self.sb[0]:
            self.socket.close()
            return # TODO: throw exception on errors

        # receive chars one by one until we reach the end sequence
        while bline[-2:] != end_seq:
            try:
                char = self.socket.recv(1)
                if not char:
                    break
                bline += char
            except socket.timeout:
                self.socket.close()
                return  # TODO: throw exception on errors

        # convert the byte stream to string and try to extract a valid HL7 message out of it
        message = self._extractor.validate(bline)
        self.socket.close()

        return message


    def send(self, message: typing.Union[bytes, bytearray, hl7.Message]):
        """ sends a message to an HL7 server.

        Args:
            message: The message may be an hl7.message or a bytearray.  sb/eb/cr are added around the message

        Returns: a string with the response (without the sb/eb/cr around the message)
        """
        if isinstance(message, hl7.Message):
            self.socket.send(self.sb + (str(message)).encode(self.encoding) + self.cr + self.eb + self.cr)
        elif isinstance(message, (bytes, bytearray)):
            self.socket.send(message)
        else:
            raise TypeError('message should be hl7.Message or a bytearray (bytes)')
        return self._receive()